package common;/*
 * LINK framework for Java 1.8
 * LINK is an experimental lightweight framework which aims to increase the likelyhood of
 * producing correct applications. This is achieved by forming computation chains,
 * inspired by Floyd-Hoare and Propositional Logic, in which resulting system can be viewed as a formalized state machine.
 * LINK supports "Throwable" execution flow, Recursion, Computation refinement,
 * Computation retries for timing and intermittent issues.

 * Copyright (C) 2013  Aran Hakki (MEng Systems Engineering, University of Warwick)

 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.


 * If you are interested in this software and have any questions please feel free to contact:

 * Aran Hakki
 * MEng Systems Engineering
 * University of Warwick

 * Email: aran.hakki@gmail.com
 * Mob: 07824098949
 * Linkedin: http://uk.linkedin.com/in/aranhakki
 */

import java.util.*;

public abstract class Machine extends Thread{

    // will apply logic when sending messages to machine instead, interrupts etc.
    protected final List<Runnable> concurrentDataConsistancyCorrectionRules = new LinkedList<Runnable>();

    public final FastPriorityActionQueue inbound;
    public Machine(int maxActionPriority){
        inbound = new FastPriorityActionQueue(maxActionPriority);
    }

    public Machine(){
        inbound = new FastPriorityActionQueue(10);
    }

    public Throwable throwable;

    private void tryExecuteAfterAssertXOR(Runnable action, Proposition... props){
        if (Config.isInTestMode){
            assertPropsAreXOR(props);
        }
        tryExecute(action);
    }

    private void tryExecute(Runnable action){
        try {
            action.run();
        } catch(Throwable e){
            e.printStackTrace();
            throwable = e;
        }
    }

    static class NotCompleteException extends RuntimeException{}
    static class NotSoundException extends RuntimeException{}

    private void assertPropsAreXOR(Proposition... props){
        int matches = 0;
        for (Proposition p : props){
            try {
                if (p.eval()){
                    matches++;
                }
            } catch (Exception e){
                continue;
            }
        }

        if (matches==0){
            throw new NotCompleteException();
        } else if (matches==1){
            //System.out.println("COMPLETE and SOUND");
        } else if (matches>1){
            throw new NotSoundException();
        }
    }

    // Action executor and priority interweave mechanism,
    // do we need this aswell as the priority interupt mechanism?
    // once the interrupt mech is implemented a new message add should,
    // interupt the currently executing transaction, rollback, start the more important action,
    // then sequence the interrupted action to start right executeAfter

    public void execute(Integer priority, Runnable action, Proposition... props){
       Action peek = inbound.peek();
        if (peek!=null){
            if (priority.compareTo(peek.priority)>=0){
                tryExecuteAfterAssertXOR(action,props);
            } else {
                Action msg = inbound.remove();
                if (msg!=null){
                    msg.action.run();
                }
            }
        } else {
            tryExecuteAfterAssertXOR(action,props);
        }
    }

    public void execute(Runnable action, Proposition... props){
        execute(0, action, props);
    }

//   // public Action current;
//    public void execute(Action a, Proposition... props){
//    //   current = a;
//        execute(a.priority, a.action, props);
//    }

    public void send(Action msg){
        inbound.add(msg);
    }

	@Override
	final public void run(){
        while(true){
            final Action action = inbound.remove();
            if (action!=null){
                Proposition p1 = ()->action.executeAfter ==null && action.executeBefore ==null;
                Proposition p2 = ()->action.executeAfter !=null && action.executeBefore ==null;
                Proposition p3 = ()->action.executeAfter ==null && action.executeBefore !=null;
                Proposition p4 = ()->action.executeAfter !=null && action.executeBefore !=null;
                assertPropsAreXOR(p1, p2, p3, p4);
                if (p1.eval()){
                    action.action.run();
                } else if (p2.eval()){
                    if (Calendar.getInstance().getTime().after(action.executeBefore)){
                        action.action.run();
                    }
                } else if (p3.eval()){
                    if (Calendar.getInstance().getTime().before(action.executeAfter)){
                        action.action.run();
                    }
                } else if (p4.eval()){
                    if (Calendar.getInstance().getTime().before(action.executeAfter) &&
                            Calendar.getInstance().getTime().after(action.executeBefore)){
                        action.action.run();
                    }
                }
            }
        }
	}

}

